import logging
from typing import Any, Optional, List, Tuple
import scipy
#import cProfile
import numpy as np
import numpy.typing as npt
from multiprocessing import Pool
from pyfairsim.sim_algorithm.OTF_Provider import (
    get_apotization_array,
    get_attenuation_array,
)
from pyfairsim.sim_algorithm.SIM_Parameters import Parameters
from pyfairsim.sim_algorithm.Wiener_Filter import Wiener_Filter
import matplotlib.pyplot as plt
from pyfairsim.utils.tif import save_tif

logger = logging.getLogger(__name__)


def plot_img(img, title):
    plt.imshow(np.real(img), cmap="gray")
    plt.title(title)
    plt.show()


def plot_fft(fft, title=""):
    plt.imshow(np.log10(np.abs(np.fft.fftshift(fft))), cmap="gray")
    plt.title(title + " in feq. region")
    plt.show()


def plot_ps(fft, peak=None, distance=None, title=""):
    """
    Function to plot the powerspectrum used for a fit and show where the peak was found and what range was excluded
    @param fft input array/image
    @param peak (kx, ky) tuple that contains the peak position if set a circle around the peak will be drawn
    @param distance that was excluded from the search if set the circle will be drawn
    @param title string to add to the title of the plot
    """
    pw = np.copy(fft)
    h, w = pw.shape
    pw = np.log10(np.abs(pw))
    # shift DC component to the center
    pw = np.fft.fftshift(pw)
    pw2 = np.copy(pw)

    if not distance == None:
        for x in range(w):
            for y in range(h):
                # check if the current position is close to the circle and if so color it white
                if (
                    np.hypot(x - w / 2, y - h / 2) < distance
                ):  # + 0.5 and np.hypot(x-w/2, y-h/2) > distance - 0.5:
                    pw2[y][x] = pw[0][0]

    mi = np.min(pw2)
    ma = np.max(pw2)
    if ma - mi > 30:
        mi = ma - 30
    pw = (pw - mi) / (ma - mi)
    ma = ma / (ma - mi)
    pw[pw < 0.8] = 0

    if not distance == None:
        for x in range(w):
            for y in range(h):
                # check if the current position is close to the circle and if so color it white
                if (
                    np.hypot(x - w / 2, y - h / 2) < distance + 0.5
                    and np.hypot(x - w / 2, y - h / 2) > distance - 0.5
                ):
                    pw[y][x] = 1

    if len(peak) == 2:
        kx = w / 2 - peak[0]  # if peak[0] > 0 else peak[0] + w
        ky = h / 2 + peak[1]  # if peak[1] < 0 else h - peak[1]
        title += f" peak at ({kx}, {ky})"
        r = 10  # radius of the circle around the peak
        for x in range(w):
            for y in range(h):
                # check if the current position is close to the peak and if so color it white
                dis = np.hypot(x - kx, y - ky)
                if dis < r + 0.5 and dis > r - 0.5:
                    pw[y][x] = 1

    # save_tif(pw, f"{os.curdir}/power_{title}_feq_region.tif")
    plt.imshow(pw, cmap="gray")
    plt.title(f"power {title} in feq. region")
    plt.show()


def get_fwhm(fwhm_fairsim: float, pixel_size: float) -> float:
    """
    Calculates the fwhm for attenuation generated with @OTF_Provider.get_attenuation_array() from the FairSIM parameters
    fwhm_fairsim: fwhm as used in FairSIM in cycles/micron
    pixel_size: pixel size in micron
    """
    return fwhm_fairsim * pixel_size


def calc_separation_matrix(
    phases: npt.ArrayLike, scalings: List[float], bands: int
) -> npt.ArrayLike:
    """
    Obtain a separation matrix for arbitrary amount of bands.
    Note: Not really tested with more than 3 bands yet.
    If #phases > #bands*2-1, pseudo-inverse will be used.
    @param phases Phases 0..2pi, order: phases[band-1][#phase]
    @param bands  Number bands (2 for 2beam, 3 for 3beam data, ..)
    @param scalings  Scaling factors, one for each band
    @return Inverted band separation matrix
    """
    number_phases = len(phases[0])
    if scalings == None:
        scalings = [1.0 if i == 0 else 0.5 for i in range(bands)]
    else:
        for index in range(1, len(scalings)):
            scalings[index] *= 0.5

    matrix = np.zeros((number_phases, bands * 2 - 1), dtype=complex)
    m, n = matrix.shape
    for phase in range(number_phases):
        # band zero is one, DC component (col 0)
        matrix[phase][0] = scalings[0]
        # all other bands: two phases per band (2 dirac peaks)
        for band in range(bands - 1):
            # pos. dirac peak (cols 1,3,5..)
            matrix[phase][2 * band + 1] = scalings[band + 1] * (
                np.cos(phases[band][phase]) + 1j * np.sin(phases[band][phase])
            )
            # neg. dirac peak (cols 2,4,6..)
            matrix[phase][2 * band + 2] = scalings[band + 1] * (
                np.cos(-phases[band][phase]) + 1j * np.sin(-phases[band][phase])
            )
    if m == n:  # calculate inverse matrix and return
        return np.linalg.inv(matrix)
    else:  # calculate pseudoinverse matrix (A+ = (A* A)^-1 A*) and return
        return (
            np.linalg.inv((matrix.conj().T @ matrix)) @ matrix.conj().T
        )  # @ is matrix multiplication


def separate_bands(
    input: npt.ArrayLike,
    phases: npt.ArrayLike,
    bands: int,
    scalings: Optional[List[float]] = None,
) -> npt.ArrayLike:
    """
    Compute the spectral separation. This creates the matrix
    first (see {@link #calc_separation_matrix}), then multiplies the spectra.
    Supports arbitrary #phases, if (#phases > bands*2-1), pseudo-inverse matrix is used.
    Phase input array should hold all (absolute) phases for 1. band,
    higher band phases are set to multiplies of base.
    @param input     FFT'd input images, array size: #images = #phases/#bands
    @param phases Phases 0..2pi, length >= (bands*2)-1
    @param bands  Number of bands, convention: 2 for 2beam, 3 for 3beam data
    @param scalings    Scaling factors, one for each band
    """
    h, w = input[0].shape
    out = np.zeros((bands * 2 - 1, h, w), dtype=np.complex64)

    # compute phases for higher bands
    pha = np.zeros((bands - 1, len(phases)))
    for band in range(1, bands):
        for phase in range(len(phases)):
            pha[band - 1][phase] = phases[phase] * band
    # create separation matrix
    separation_matrix = calc_separation_matrix(pha, scalings, bands)
    # multiply, output 0 .. bands*2-1
    for phase in range(len(phases)):
        for band in range(bands * 2 - 1):
            out[band] += separation_matrix[band][phase] * input[phase]

    return out


def locate_peak(crosscorrelation: npt.ArrayLike, min_distance: float) -> Tuple[float]:
    """
    Locates position, magnitute and phase of the highest peak in 'crosscorrelation'.
    @param crosscorrelation input vector (typ. cross-/auto-correlation)
    @param min_distance Mininum distance from DC component, in pxl
    @return px, py, mag, phase
    """
    h, w = crosscorrelation.shape

    x_pos, y_pos, maximum, phase = -1, 1, np.min(np.abs(crosscorrelation)), 0.0
    # since if we dont find a max we negate the y we should set it to 1 so this returns -1

    for y in range(h):
        for x in range(w):
            # distance to DC component
            rad = np.sqrt(
                (x ** 2 if x < w / 2 else (x - w) ** 2)
                + (y ** 2 if y < h / 2 else (y - h) ** 2)
            )
            if rad > min_distance and np.abs(crosscorrelation[y][x]) > maximum:
                maximum = np.abs(crosscorrelation[y][x])
                x_pos = x
                y_pos = y
                phase = np.angle(crosscorrelation[y][x])

    # convert to our coordinate convention
    if x_pos > w / 2:
        x_pos = x_pos - w
    if y_pos > h / 2:
        y_pos = h - y_pos
    else:
        y_pos *= -1

    return x_pos, y_pos, maximum, phase


def fit_peak(
    band_0: npt.ArrayLike,
    band_1: npt.ArrayLike,
    band_0_index: int,
    band_1_index: int,
    parameters: Parameters,
    kx: float,
    ky: float,
    weight_limit: float,
    search: float,
) -> Tuple[Any]:
    """
    Fits SIM parameters by cross-correlation of common frequency components.
    Correlates band0 vs. band1, with band1 shifted to kx,ky.

    @param band_0 Low freq band to correlate to (that does not change position)
    @param band_1 High freq band to correlate
    @param band_0_index  Number of the low band (usually 0)
    @param band_1_index  Number of the high band (usually 1 or 2)
    @param cutoff cutoff for the otf
    @param kx Starting guess kx
    @param ky Starting guess ky
    @param weight_Limit If > 0, consider only freq where both weights are over this limit
    @param search kx,ky will be varied +-search
    @return [px, py, mag, phase] and cntrl control vector (size 30x10) to be filled with power spectra, may be null
    """
    # ToDo: check if all arrays are same size

    result_phase = 0
    result_mag = 0
    control = np.zeros((10, 30))

    # loop iterations, each iteration does closer search
    for iteration in range(3):
        # copy input data
        b0 = np.copy(band_0)
        b1 = np.copy(band_1)
        # define common region, with current search guess
        common_region(
            b0,
            b1,
            band_0_index,
            band_1_index,
            parameters,
            kx,
            ky,
            0.15,
            weight_limit,
            True,
        )  # ToDo hardcoded parameters
        # go to real space
        b0 = np.fft.ifft2(b0)
        b1 = np.fft.ifft2(b1)

        # store all correlations
        correlations = np.zeros((10, 10), dtype=complex)
        scale = 1.0 / np.sum(np.abs(b0) ** 2)

        tkx = kx
        tky = ky
        ts = search

        for xi in range(10):
            for yi in range(10):
                # compute position to shift to
                x_pos = tkx + ((xi - 4.5) / 4.5) * ts
                y_pos = tky + ((yi - 4.5) / 4.5) * ts

                # copy and fourier-shift band 1
                b1s = np.copy(b1)
                b1s = fourier_shift(b1s, x_pos, -y_pos)

                # get correlation by multiplication, suming elements, scaling by b0
                correlations[yi][xi] = (
                    np.sum(np.conjugate(b0) * b1s) * scale
                )  # ToDo check if this is correct

        # find the maximum, set as new starting point
        maximum = 0
        minimum = (
            np.max(np.abs(correlations)) + 1
        )  # to ensure it is larger than any that can occur
        new_kx = 0
        new_ky = 0
        for yi in range(10):
            for xi in range(10):
                absolute = np.abs(correlations[yi][xi])
                if absolute > maximum:
                    maximum = absolute
                    new_kx = tkx + ((xi - 4.5) / 4.5) * ts
                    new_ky = tky + ((yi - 4.5) / 4.5) * ts
                    result_phase = np.angle(correlations[yi][xi])
                    result_mag = absolute
                if absolute < minimum:
                    minimum = absolute

        # output to control vector
        for yi in range(10):
            for xi in range(10):
                control[yi][xi + iteration * 10] = (
                    np.abs(correlations[yi][xi]) - minimum
                ) / (maximum - minimum)

        # set new guess
        kx = new_kx
        ky = new_ky
        search /= 5

    return ([kx, ky, result_phase, result_mag], control)


def fourier_shift(array: npt.ArrayLike, kx, ky) -> npt.ArrayLike:
    N, _ = array.shape  # Assumes square array
    x = np.outer(np.ones(N), np.array(range(N)))
    y = np.outer(np.array(range(N)), np.ones(N))
    phase_shift = np.exp(1j * 2 * np.pi * (kx * x + ky * y) / N)
    array *= phase_shift
    return array


def common_region(
    band_0: npt.ArrayLike,
    band_1: npt.ArrayLike,
    band_0_index: int,
    band_1_index: int,
    parameters: Parameters,
    kx: float,
    ky: float,
    distance: float,
    weight_limit: float,
    divide_by_otf: bool,
) -> None:
    """
    Determines a common freq region of band0, band1.
    All frequencies where either OTF is below threshhold 'weightLimit'
    are set to zero in both bands. All others are divided by weight (if switch is set).
    @param band0 The lower band, will not be moved
    @param band1 The higher band, shifted to kx,ky
    @param bn0  Number of the low band (usually 0)
    @param bn1  Number of the high band (usually 1 or 2)
    @param cutoff cutoff for the otf
    @param kx x-coordinate of shift
    @param ky y-coordinate of shift
    @param dist Minimal distance (as fraction of hypot(kx,ky)) to band centers (std use 0.15)
    @param weightLimit Components where either OTF is below this limit are set to 0
    @param divideByOtf If set, all components over weightLimit are divided by OTF weight
    """
    # ToDo: check if all arrays are same size
    shape = band_0.shape
    h, w = shape
    weight_0 = parameters.get_otf(attenuated=False)
    weight_1 = parameters.get_otf(attenuated=False)
    wt_0 = parameters.get_otf(kx=kx, ky=ky, attenuated=False)
    wt_1 = parameters.get_otf(kx=-kx, ky=-ky, attenuated=False)

    # set zero if minimal weight not reached in one or both OTFs
    band_0[np.abs(weight_0) < weight_limit] = 0
    band_0[np.abs(wt_0) < weight_limit] = 0
    if divide_by_otf:
        with np.errstate(
            invalid="ignore", divide="ignore"
        ):  # prevents the runtimeWarning to be printed
            band_0 /= weight_0

    band_1[np.abs(weight_1) < weight_limit] = 0
    band_1[np.abs(wt_1) < weight_limit] = 0
    if divide_by_otf:
        with np.errstate(invalid="ignore", divide="ignore"):
            band_1 /= weight_1

    # set zero around DC component
    maximum = np.sqrt(kx ** 2 + ky ** 2)
    y, x = np.ogrid[0 : shape[0], 0 : shape[1]]
    distances = np.sqrt((x - (w / 2)) ** 2 + (y - (h / 2)) ** 2)
    with np.errstate(invalid="ignore", divide="ignore"):
        relative_distances = np.fft.ifftshift(distances) / maximum
    band_0[relative_distances < distance] = 0
    band_0[relative_distances > (1 - distance)] = 0

    shifted_distances = np.sqrt((x - (w / 2) + kx) ** 2 + (y - (h / 2) - ky) ** 2)
    with np.errstate(invalid="ignore", divide="ignore"):
        relative_shifted_distances = np.fft.ifftshift(shifted_distances) / maximum
    band_1[relative_shifted_distances < distance] = 0
    band_1[relative_shifted_distances > (1 - distance)] = 0

    # fix divide by zeros, we want to only divide when the weight is larger then weight_limit but since we else set it to 0 we might aswell divide all and fix all the divide by zeros
    band_0[np.isnan(band_0)] = 0
    band_1[np.isnan(band_1)] = 0


def get_peak(
    band0: npt.ArrayLike,
    band1: npt.ArrayLike,
    band_0_index: int,
    band_1_index: int,
    parameters: Parameters,
    kx: float,
    ky: float,
    weight_limit: float,
) -> complex:
    """
    Get the cross-correlation of two bands. This fourier-shifts b1 to kx,ky, and
    only takes into account regions where both OTFs are over weightLimit
    (Also, a minimum distance of 0.15*hypot(kx,ky) is ensured to the bands centers).
    Result is scaled by |band0|^2, thus might be used as estimate for modulation.
    @param band0 The lower band, will not be moved
    @param band1 The higher band, shifted to kx,ky
    @param bn0  Number of the low band (usually 0)
    @param bn1  Number of the high band (usually 1 or 2)
    @param cutoff cutoff for the otf
    @param kx x-coordinate of shift
    @param ky y-coordinate of shift
    @param weightLimit Limit for taking into account a certain frequence
    """
    # copy input data
    b0 = np.copy(band0)
    b1 = np.copy(band1)
    # define common freq. region
    common_region(
        b0, b1, band_0_index, band_1_index, parameters, kx, ky, 0.15, weight_limit, True
    )
    # plot_ps(b1, "common")

    # go to real space
    b0 = np.fft.ifft2(b0)
    b1 = np.fft.ifft2(b1)
    # plot_img(b0, "b0")
    # plot_img(b1, "b1")

    # Fourier-shift band1 to correct position
    b1 = fourier_shift(b1, kx, -ky)

    # element wise mult of b0 conjugate with b1
    b1 *= np.conjugate(b0)
    # sum, scale by |band2|^2 and return
    return np.sum(b1) / np.sum(np.abs(b0) ** 2)


def paste_freq(
    array: npt.ArrayLike, out_shape: Tuple[int], x_offset: int, y_offset: int
) -> npt.ArrayLike:
    ret = np.zeros(out_shape, dtype=np.complex128)

    hi, wi = array.shape
    ho, wo = out_shape

    for x in range(wi):
        for y in range(hi):
            xo = x if x < wi / 2 else x + wo // 2
            yo = y if y < hi / 2 else y + ho // 2
            xo = (xo + x_offset + wo) % wo
            yo = (yo + y_offset + ho) % ho
            ret[yo][xo] = array[y][x]

    return ret


def estimate_absolute_phases(parameters: Parameters, input_FFT: npt.ArrayLike):
    for angle in range(len(parameters.directions)):
        direction = parameters.directions[angle]
        phases = []
        for i in range(direction.number_phases):
            ac = np.copy(input_FFT[angle][i])
            # ac *= np.conjugate(parameters.otf_attenuated_array)
            correlation = auto_correlation(ac, direction.px, direction.py)
            phases.append(np.angle(correlation))
        direction.phases = np.array(phases)
        logger.info(f"Found phases for {angle} : {phases}")
        direction.has_individual_phases = True
        direction.set_phase_offset(0)


def estimate_absolute_phases_parallel(parameters: Parameters, input_FFT: npt.ArrayLike):
    arg_list = []
    # plot_img(parameters.otf_attenuated_array, "otf_att")
    # plot_img(parameters.otf_array, "otf")
    # return
    for angel_index in range(len(parameters.directions)):
        current_dir = parameters.directions[angel_index]
        for phase in range(current_dir.number_phases):
            arg_list.append(
                (
                    np.copy(input_FFT[angel_index][phase]),
                    current_dir.px,
                    current_dir.py,
                    parameters.get_otf(attenuated=False),
                )
            )
    with Pool(len(parameters.directions) * parameters.directions[0].number_phases) as p:
        results = p.starmap(estimate_absolute_phase, arg_list)
    for angel_index in range(len(parameters.directions)):
        current_dir = parameters.directions[angel_index]
        current_dir.set_phase_offset(0)
        current_dir.has_individual_phases = True
        current_dir.phases = np.array(
            results[
                parameters.directions[0].number_phases
                * angel_index : parameters.directions[0].number_phases
                * angel_index
                + parameters.directions[0].number_phases
            ]
        )
        logger.info(f"Found phases for {angel_index} : {current_dir.phases}")


def estimate_absolute_phase(fft_image, px, py, otf):
    fft_image *= np.conjugate(otf)
    return np.angle(auto_correlation(fft_image, px, py))


def auto_correlation(input: npt.ArrayLike, kx: float, ky: float) -> np.complex128:
    # double the vector size to allow a good shift
    aV = paste_freq(input, (input.shape[0] * 2, input.shape[1] * 2), 0, 0)
    bV = paste_freq(input, (input.shape[0] * 2, input.shape[1] * 2), 0, 0)
    # ToDo need to check if the shift + otf support still fit in the array if so we can use the original else we need to magnify it
    # aV = np.copy(input)
    # bV = np.copy(input)
    # plot_fft(aV, "av")
    # plot_fft(bV, "bv")
    # move one copy to its new position kx, ky
    bV = np.fft.fft2(fourier_shift(np.fft.ifft2(bV), kx, ky))

    # compute the auto-correlation
    product = aV * np.conjugate(bV)
    ret = np.sum(product)

    # normalize (?)
    ret = ret / np.sum(aV * aV)
    return ret


def estimate_parameters_one_direction(
    angel_index: int,
    parameters: Parameters,
    input_FFT: npt.ArrayLike,
    otf_attenuation: npt.ArrayLike,
    keep_phases: bool,
):
    #profiler = cProfile.Profile()
    #profiler.enable()
    current_dir = parameters.directions[angel_index]
    h, w = input_FFT[0][0].shape
    fit_band = parameters.sim_parameters.fit_band

    # idx of low band (phase detection) and high band (shift vector detection)
    # will be the same for two-beam
    low_band = 1
    high_band = 3 if current_dir.number_bands == 3 else 1
    fb = low_band if fit_band == 1 else high_band

    # compute band seperation
    input_copy = np.zeros((current_dir.number_phases, h, w), dtype=complex)
    separate = np.zeros((current_dir.get_number_of_components(), h, w))

    for phase in range(current_dir.number_phases):
        input_copy[phase] = np.copy(input_FFT[angel_index][phase])
        input_copy[phase] *= current_dir.get_intensity_quotient(phase)

    if not keep_phases:
        # reset any previously set phases
        current_dir.reset_phases()
    else:
        raise NotImplementedError("keep_phases not implemented jet")  # ToDo

    phases = current_dir.get_phases()
    phases = np.array(phases)

    separate = separate_bands(input_copy, phases, current_dir.number_bands)
    # ToDo is the line pattern in the difference between this and fairSIM, see file on desktop, relevant?
    # save_tif(separate[0].real, f"separate{angel_index}0.tif")
    # save_tif(separate[1].real, f"separate{angel_index}1.tif")
    # save_tif(separate[2].real, f"separate{angel_index}2.tif")

    # compute correlation: dampen region around DC, ifft, mult. in spatial, fft back
    # save_tif(otf_attenuation, "otfattenuated.tif")
    c_0 = np.fft.ifft2(np.copy(separate[0]) * otf_attenuation)
    c_1 = np.fft.ifft2(np.copy(separate[low_band]) * otf_attenuation)
    c_2 = np.fft.ifft2(np.copy(separate[high_band]) * otf_attenuation)

    c_1 *= np.conjugate(c_0)
    c_2 *= np.conjugate(c_0)

    c_1 = np.fft.fft2(c_1)
    c_2 = np.fft.fft2(c_2)
    # save_tif(c_1.real, "c_1.tif")
    # ToDo same pattern still in difference at 0.03% of the max amplitude of original

    min_distance = 2

    if parameters.sim_parameters.fit_exclude > 0:
        # find the highest peak in corr of band0 to highest band
        # minDist of otfCutoff from origin, store in 'param'
        min_distance = (
            parameters.sim_parameters.fit_exclude
            * parameters.cutoff
            * parameters.image_parameters.image_shape[0]
        )
        peak = locate_peak(c_1 if fit_band == 1 else c_2, min_distance)
    else:
        raise NotImplementedError("fit_exclude not implemented jet")  # ToDo

    # fit the peak to sub-pixel precision by cross-correlation of the Fourier-shifted components
    peak, control = fit_peak(
        separate[0],
        separate[fb],
        0,
        fb,
        parameters,
        -peak[0],
        -peak[1],
        0.05,
        2.5,
    )
    # plot_ps(c_1 if fit_band == 1 else c_2, peak=peak[:2], distance=min_distance)
    # plot_img(control, "control")

    # now, either three beam / 3 bands ...
    if low_band != high_band:
        # peak should contain the shift band0 <-> band2, so if band0 <-> band1 was fitted, multiply by 2
        if fit_band == 1:
            peak[0] *= 2
            peak[1] *= 2

        # at the peak position found, extract phase and modulation from band0 <-> band1
        peak_1 = get_peak(
            separate[0],
            separate[low_band],
            0,
            1,
            parameters,
            peak[0] / 2,
            peak[1] / 2,
            0.05,
        )

        # TODO: this is a quick fix, this should be done properly by looking at the OTF
        # overlap, to figure out if parameter extraction is possible. This just catches
        # the 'definitely not possible' case
        if np.hypot(peak[0], peak[1]) > (
            parameters.cutoff * parameters.image_parameters.image_shape[0]
        ):
            peak_2 = peak_1
        else:
            # extract modulation from band0 <-> band2
            peak_2 = get_peak(
                separate[0],
                separate[high_band],
                0,
                2,
                parameters,
                peak[0],
                peak[1],
                0.05,
            )

        # return the result
        #       px        py       phase_offset      modulation[1]   modulation[2]
        return (-peak[0], peak[1], np.angle(peak_1), np.abs(peak_1), np.abs(peak_2))
    # ... or two-beam / 2 bands
    if low_band == high_band:
        # get everything from one correlation band0 <-> band1
        peak_1 = get_peak(
            separate[0], separate[1], 0, 1, parameters, peak[0], peak[1], 0.05
        )

        #profiler.disable()
        #if angel_index == 0:
        #    profiler.print_stats(sort="tottime")

        # return the result
        #       px        py       phase_offset      modulation[1]   modulation[2]
        return (-peak[0], peak[1], np.angle(peak_1), np.abs(peak_1), -1)


def estimate_parameters(
    parameters: Parameters,
    input_FFT: npt.ArrayLike,
    keep_phases: bool,
) -> None:
    """
    Run the SIM parameter estimation
    @param parameters  The SIM parameter instance to work on
    @param input_FFT  The input images (in Fourier space)
    @param keepPhases If true, phase information from SimParam will be used in band separation
    """
    # input_FFT consists of 4D array 1st and 2nd dimension are band and phase, last two define image dimensions
    fit_band = parameters.sim_parameters.fit_band
    if fit_band != 1 and fit_band != 2:
        raise ValueError("Fitband neither 1 nor 2!")

    shape = input_FFT[0][0].shape

    # The attenuation vector helps well to fade out the DC component,
    # which is uninteresting for the correlation anyway
    attenuation = get_attenuation_array(
        shape, 0.99, 0.15 * parameters.cutoff
    )  # ToDo Hardcoded
    # plot_img(otf_attenuation, "att")
    # plot_img(parameters.get_otf(attenuated=False), "otf")
    # save_tif(np.fft.fftshift(parameters.get_otf(attenuated=False)), "otf.tif")
    arg_list = []
    for angel_index in range(len(parameters.directions)):
        arg_list.append((angel_index, parameters, input_FFT, attenuation, keep_phases))
    with Pool(3) as p:
        results = p.starmap(estimate_parameters_one_direction, arg_list)
    for angel_index, result in enumerate(results):
        px, py, phase_offset, modulation_1, modulation_2 = result
        current_dir = parameters.directions[angel_index]
        current_dir.set_px_py(px, py)
        current_dir.set_phase_offset(phase_offset)
        current_dir.modulations[1] = modulation_1
        if modulation_2 != -1:
            current_dir.modulations[2] = modulation_2
        if py == 0:
            py = 1e-14
        logger.info(
            f"Found {angel_index} peak with shift: ({px:.3f},{py:.3f}), r: {np.hypot(px, py):.3f}, angle: {np.arctan(px/py):.3f} and phase offset {phase_offset:.5f} and modulation {modulation_1:.5f}, {modulation_2:.3f}."
        )


def run_reconstruction(
    parameters: Parameters, inFFT: npt.ArrayLike
) -> List[npt.ArrayLike]:
    """
    Run the SIM reconstruction
    @param parameters  The SIM parameter instance to work on
    @param inFFT  The input images (in Fourier space)
    @param otf_before_shift Apply the OTF before shifting bands
    @param image_clip_scale Clip zero values and scale (0..255) output images?
    @return widefield_result The widefield image computed during the reconstruction (may be null)
    @return filtered_widefield_result The filtered widefield image (may be null)
    @return The reconstructed image */
    """
    h, w = inFFT[0][0].shape
    apo_bend = parameters.filter_parameters.apodization_bend
    apo_cutoff = parameters.filter_parameters.apodization_cutoff
    # setup Wienerfilter
    w_filter = Wiener_Filter(parameters)
    # vector to store the result
    size = parameters.image_parameters.image_shape[0] * 2
    full_result = np.zeros((size, size), dtype=complex)
    low_frequency_result = np.zeros((size, size), dtype=complex)
    # ToDo here RL stuff
    # loop all pattern directions
    for angle_index in range(len(parameters.directions)):
        current_dir = parameters.directions[angle_index]
        separate = [np.zeros((h, w)) for _ in range(current_dir.number_bands * 2 - 1)]
        # copy into temp. array (to not override input data) and apply correction factor
        tmp_array = [np.zeros((h, w)) for _ in range(current_dir.number_phases)]
        for phase in range(current_dir.number_phases):
            tmp_array[phase] = np.copy(inFFT[angle_index][phase])
            tmp_array[phase] *= current_dir.get_intensity_quotient(phase)

        # ToDo here RL stuff
        # use the temp array as input for the band separation
        separate = separate_bands(
            tmp_array,
            current_dir.get_phases(),
            current_dir.number_bands,
            current_dir.get_modulations(),
        )
        # add band 0 to low_frequency_result
        low_frequency_result += paste_freq(separate[0], (2 * h, 2 * w), 0, 0)

        for i in range(current_dir.number_bands * 2 - 1):
            # plot_fft(shifted[i], f"shifted[{i}]")
            # plot_img(np.fft.ifft2(separate[i]), i)
            pass
        # Wiener filter: Apply OTF here
        if (
            parameters.sim_parameters.otf_before_shift and True
        ):  # ToDo True should be replaced by parameter.use_wiener_filter() for now hardcoded wiener filter use
            otf_array = parameters.get_otf()
            for i in range(current_dir.number_bands * 2 - 1):
                # plot_fft(separate[i], f"i separate[{i}], angle: {angle_index}")
                separate[i] *= np.conj(otf_array)
                # plot_fft(separate[i], f"f separate[{i}], angle: {angle_index}")
        # Shifts to correct position
        shifted = [
            np.zeros((2 * h, 2 * w)) for _ in range(5)
        ]  # ToDo why 5 why hardcoded
        # band0 is DC, so does not need shifting, only a bigger vector
        shifted[0] = paste_freq(separate[0], (2 * h, 2 * w), 0, 0)
        # higher bands need shifting
        for band in range(1, current_dir.number_bands):
            positive = band * 2
            negative = band * 2 - 1
            # first copy to larger vectors
            shifted[positive] = paste_freq(separate[positive], (2 * h, 2 * w), 0, 0)
            shifted[negative] = paste_freq(separate[negative], (2 * h, 2 * w), 0, 0)
            # then fourier shift
            shifted[positive] = np.fft.fft2(
                fourier_shift(
                    np.fft.ifft2(shifted[positive]),
                    current_dir.px * band,
                    current_dir.py * band,
                )
            )
            shifted[negative] = np.fft.fft2(
                fourier_shift(
                    np.fft.ifft2(shifted[negative]),
                    -current_dir.px * band,
                    -current_dir.py * band,
                )
            )
        if (
            True
        ):  # ToDo True should be replaced by parameter.use_wiener_filter() for now hardcoded wiener filter use

            if not parameters.sim_parameters.otf_before_shift:
                # multiply with shifted OTF
                otf_array = np.zeros((2 * h, 2 * w), dtype=np.complex128)
                otf_array[
                    2 * h // 4 : (3 * 2 * h) // 4, 2 * w // 4 : (3 * 2 * w) // 4
                ] += np.fft.ifftshift(parameters.get_otf())

                shifted[0] *= otf_array
                for band in range(1, current_dir.number_bands):

                    otf_array_plus = scipy.ndimage.shift(
                        otf_array, (-current_dir.py * band, current_dir.px * band)
                    )
                    otf_array_plus = np.fft.ifftshift(otf_array_plus)
                    otf_array_minus = scipy.ndimage.shift(
                        otf_array, (current_dir.py * band, -current_dir.px * band)
                    )
                    otf_array_minus = np.fft.ifftshift(otf_array_minus)
                    positive = band * 2
                    negative = band * 2 - 1
                    shifted[positive] *= otf_array_plus
                    shifted[negative] *= otf_array_minus
            else:
                otf_array = np.zeros((2 * h, 2 * w), dtype=np.complex128)
                otf_array[
                    2 * h // 4 : (3 * 2 * h) // 4, 2 * w // 4 : (3 * 2 * w) // 4
                ] += np.fft.ifftshift(parameters.get_otf(attenuated=False))
                # or mask for OTF support
                for band in range(1, current_dir.number_bands):
                    positive = band * 2
                    negative = band * 2 - 1
                    otf_array_plus = scipy.ndimage.shift(
                        otf_array, (current_dir.py * band, current_dir.px * band)
                    )
                    otf_array_plus = np.fft.ifftshift(otf_array_plus)
                    otf_array_minus = scipy.ndimage.shift(
                        otf_array, (-current_dir.py * band, -current_dir.px * band)
                    )
                    otf_array_minus = np.fft.ifftshift(otf_array_minus)
                    shifted[positive][otf_array_plus < 1e-7] = 0
                    shifted[negative][otf_array_minus < 1e-7] = 0

        # sum up result
        # ToDo RL
        for i in range(current_dir.number_bands * 2 - 1):
            full_result += shifted[i]
            # plot_fft(shifted[i], f"shifted[{i}]")
            # plot_img(np.fft.ifft2(shifted[i]), i)

        for i in range(current_dir.number_bands):
            band_image = shifted[i * 2]
            if i != 0:
                band_image += shifted[i * 2 - 1]
            # plot_fft(band_image, f"a{angle_index}: band:{i}")
            # plot_img(np.fft.ifft2(band_image), f"a{angle_index}: band:{i}")
            # save_tif(
            #    np.fft.ifft2(band_image).astype(np.float32),
            #    f"shifted{angle_index}{i}.tif",
            # )
    # -- done loop all pattern directions, 'fullResult' now holds the image --

    # Wiener filter on output

    if (
        True
    ):  # ToDo True should be replaced by parameter.use_wiener_filter() for now hardcoded wiener filter use
        # multiply by wiener denominator
        w_den = w_filter.get_denominator(
            parameters.filter_parameters.wiener_filter_parameter
        )
        widefield_wiener_denominator = w_filter.get_widefield_denominator(
            parameters.filter_parameters.wiener_filter_parameter
        )
        # wiener = 1/w_den
        # wiener -= np.min(wiener)
        # wiener /= np.max(wiener)
        # wiener = np.fft.fftshift(wiener)
        # plot_img(wiener, f"w")
        # save_tif(wiener.astype(np.float32), "py_wiener.tif")
        full_result *= w_den

        otf_array = np.zeros_like(low_frequency_result)
        otf_array[
            2 * h // 4 : (3 * 2 * h) // 4, 2 * w // 4 : (3 * 2 * w) // 4
        ] += np.fft.ifftshift(parameters.get_otf(attenuated=False))
        otf_array = np.fft.ifftshift(otf_array)
        low_frequency_result[otf_array < 0.001] = 0
        low_frequency_result *= otf_array
        low_frequency_result *= widefield_wiener_denominator
        # plot_img(w_den, "wfd")
        # save_tif(widefield_wiener_denominator.real, "wfd.tif")

        # apply apotization filter
        # apo = np.zeros((2 * h, 2 * w), dtype=complex)
        # otf.write_apo_vector(apo, apo_bend, apo_cutoff)
        apo = get_apotization_array(
            (2 * h, 2 * w), 1 / 2 * apo_cutoff * parameters.cutoff, apo_bend
        )
        # apo_real = np.fft.fftshift(apo)
        # plot_img(apo_real, f"w")
        # save_tif(apo_real, "apo.tif")
        full_result *= apo
        # plot_fft(full_result, f"full result")

        full_result_image = np.real(
            np.fft.ifft2(full_result)
        )  # ToDo this would need cliping and scaling
        filtered_widefield_image = np.real(np.fft.ifft2(low_frequency_result))

    # ToDo RL

    # print(np.min(full_result_image))
    return [full_result_image, filtered_widefield_image]  # ToDo add widefield



from pyfairsim.utils.load_settings import create_settings_file
from pyfairsim.utils.argparse import parse_args_create_settings_file

if __name__ == "__main__":
    args = parse_args_create_settings_file()
    filename = (
        args.filename if args.filename.endswith(".json") else (args.filename + ".json")
    )
    create_settings_file(filename)



import logging
import time

import numpy as np

from pyfairsim.sim_algorithm.SIM_Utils import estimate_parameters

from pyfairsim.utils.argparse import parse_args_k_estimation
from pyfairsim.utils.load_settings import load_settings, save_settings
from pyfairsim.utils.tif import load_tif, order_images, fft_images


def main():
    args = parse_args_k_estimation()

    logging.basicConfig(level=args.level)
    global logger
    logger = logging.getLogger()

    print(f"Running parameter estimation for {args.filename}")

    if args.settings_filename:
        logger.debug(f"loading settings from {args.settings_filename}")
        settings = load_settings(args.settings_filename)
    else:
        logger.warning(f"no settings provided fall back on default settings")
        settings = load_settings(args.filename)

    logger.debug(f"loading images from {args.filename}")
    images, _ = load_tif(args.filename)

    if len(images) < len(settings.directions) * settings.directions[0].number_phases:
        logger.error(
            f"given image stack has to few images n={len(images)} but at least {len(settings.directions) * settings.directions[0].number_phases} are expected"
        )
        return

    logger.debug(f"order images")
    ordered_images = order_images(
        images,
        len(settings.directions),
        settings.directions[0].number_phases,
        settings.image_parameters.ordering,
        settings.image_parameters.z_slice,
    )

    logger.debug("fft images")
    raw_images = fft_images(
        ordered_images,
        len(settings.directions),
        settings.directions[0].number_phases,
        settings.image_parameters.background,
    )

    logger.info(f"run k-vector estimation")
    start = time.time()
    estimate_parameters(settings, raw_images, False)
    stop = time.time()
    logger.info(f"parameter estimation done in {stop-start:.2f} s.")

    if args.out_filename:
        save_settings(settings, args.out_filename, overwrite=True)
    else:
        for direction in settings.directions:
            px, py = direction.px * (
                settings.directions[0].number_bands - 1
            ), direction.py * (settings.directions[0].number_bands - 1)
            if py == 0:  # prevent ZeroDivision error
                py = 1e-14
            print(
                f"kx: {px:.3f}, ky: {py:.3f}, r: {np.hypot(px, py):.3f}, angle: {np.arctan(px/py):.3f}"
            )


if __name__ == "__main__":
    main()



import unittest
import numpy as np
from pyfairsim.sim_algorithm.SIM_Utils import locate_peak

def get_outside_position(distance, shape):
    if distance/2 > np.sqrt(shape[0]**2 + shape[1]**2):  # there is no outside position
        return -1, -1
    return shape[1]//2, shape[0]//2  # choose the outermost position

class TestSIM_Utils(unittest.TestCase):
    
    image_shapes = [[512, 512], [512, 1024], [1024, 512], [2048, 2048], [401, 100]]

    def test_locate_peak(self):
        relative_distances = (0, 0.2, 0.9, 1)
        for image_shape in self.image_shapes:
            test_array = np.zeros(image_shape, dtype=complex)
            for relative_distance in relative_distances:
                min_distance = int(0.5 * relative_distance * np.sqrt(image_shape[0]**2 + image_shape[1]**2))
                self.assertEqual(locate_peak(test_array, min_distance), (-1, -1, 0., 0.))
                test_array_with_max = np.copy(test_array)
                x, y = get_outside_position(min_distance, image_shape)
                if x != -1:
                    test_array_with_max[y][x] = 3 + 4j
                    self.assertEqual(locate_peak(test_array_with_max, min_distance), (x, -y, 5, 0.9272952180016122))

                test_array_with_max_in_radius = np.copy(test_array)
                test_array_with_max_in_radius[0][0] = 1
                self.assertEqual(locate_peak(test_array, min_distance), (-1, -1, 0., 0.))

if __name__ == "__main__":
    unittest.main()


import logging
from typing import Optional, List

import numpy as np
import numpy.typing as npt

logger = logging.getLogger(__name__)


class Direction:
    def __init__(self, number_bands: int, number_phases: int) -> None:
        if number_bands < 2:
            raise RuntimeError(f"bands ({number_bands}) < 2, not usefull")
        if number_phases < (number_bands * 2 - 1):
            raise RuntimeError(
                f"not enough phases for bands: {number_bands} {number_phases}"
            )
        self.number_bands = number_bands  # how many bands
        self.number_phases = number_phases  # how many phases
        self.px, self.py = -1, -1  # shift vector (band1)
        self.phase_offset = 0  # global phase offsets ToDo
        self.phases = np.zeros(number_phases)  # phases
        self.modulations = np.ones(number_bands)  # modulation
        self.has_individual_phases = False  # if non-equidist. phases are set
        self.angle_intensity_factor = 1  # the intensity factor to apply to this angle
        self.phase_intensity_factors = np.ones(
            number_phases
        )  # the intensity factors to apply phase-by-phase
        self.reset_phases(1)  # inits the phases to equidistant
        self.mod_low_limit = 0.4
        self.mod_high_limit = 1.1
        self.default_modulation = 0.65

    def reset_phases(self, multiplier: Optional[float] = 1.0) -> None:
        """
        Reset individual phases back to equidistant.
        """
        for i in range(self.number_phases):
            self.phases[i] = (2 * np.pi * multiplier / self.number_phases) * i
            # if self.phases[i] > np.pi:
            #    self.phases[i] -= 2 * np.pi
        self.has_individual_phases = False

    def set_px_py(self, in_px: float, in_py: float) -> None:
        self.px = in_px / (self.number_bands - 1)
        self.py = in_py / (self.number_bands - 1)

    def set_phase_offset(self, phase: float):
        self.reset_phases()
        self.phase_offset = phase

    def get_intensity_quotient(self, phase: int) -> float:
        """
        Returns a combined correction quotient: 1/(angleIntesity*phaseIntensity)
        """
        return 1 / (self.angle_intensity_factor * self.phase_intensity_factors[phase])

    def get_number_of_components(self) -> int:
        return self.number_bands * 2 - 1

    def get_phases(self) -> npt.ArrayLike:
        ret = []
        for i in range(self.number_phases):
            ret.append(self.phases[i] + self.phase_offset)
        return ret

    def get_modulations(self) -> List[float]:
        ret = [x for x in self.modulations]
        for band in range(self.number_bands):
            if ret[band] < self.mod_low_limit:
                logger.warning(
                    f"Modulation {ret[band]} below limit {self.mod_low_limit}! Modulation was set to default {self.default_modulation} instead"
                )
                ret[band] = self.default_modulation
            if ret[band] > self.mod_high_limit:
                logger.warning(
                    f"Modulation {ret[band]} above limit {self.mod_high_limit}! Modulation was set to default {self.default_modulation} instead"
                )
                ret[band] = self.default_modulation
        return ret

    def as_dict(self) -> dict:
        out_dict = {}
        out_dict["number_bands"] = self.number_bands
        out_dict["number_phases"] = self.number_phases
        out_dict["px"] = self.px
        out_dict["py"] = self.py
        out_dict["phases"] = self.get_phases()
        out_dict["modulations"] = self.get_modulations()
        out_dict["angle_intensity_factor"] = self.angle_intensity_factor
        out_dict["phase_intensity_factors"] = self.phase_intensity_factors

        return out_dict

    @classmethod
    def as_descriptive_dict(cls) -> dict:
        out_dict = {}
        out_dict["number_bands"] = "Number of bands"
        out_dict["number_phases"] = "Number of phases in the image data"
        out_dict["px"] = "x component of the shift vector"
        out_dict["py"] = "y component of the shift vector"
        out_dict["phases"] = "Phase offsets of the different phases (images)"
        out_dict["modulations"] = "Modulations"
        out_dict[
            "angle_intensity_factor"
        ] = "Factor to multiply the input data from this angel"
        out_dict[
            "phase_intensity_factors"
        ] = "Factors to multiply the different phases with"

        return out_dict

    @classmethod
    def from_dict(cls, data_dict: dict) -> "Direction":
        direction = Direction(data_dict["number_bands"], data_dict["number_phases"])
        direction.px = data_dict["px"]
        direction.py = data_dict["py"]
        direction.phases = data_dict["phases"]
        direction.modulations = data_dict["modulations"]
        direction.angle_intensity_factor = data_dict["angle_intensity_factor"]
        direction.phase_intensity_factors = data_dict["phase_intensity_factors"]

        return direction



import logging
from typing import Optional, List, Tuple
from numpy.typing import ArrayLike

from pyfairsim.sim_algorithm.OTF_Provider import (
    get_ideal_otf_array,
    get_attenuation_array,
    get_otf_from_file,
    shift_otf,
)
from pyfairsim.sim_algorithm.Direction import Direction
from pyfairsim.utils.physics import to_SI

logger = logging.getLogger(__name__)


class SIM_Parameters:
    def __init__(
        self,
        otf_correction: float,
        otf_type: str,
        otf_before_shift: bool,
        attenuation_strength: float,
        attenuation_fwhm: float,
        use_attenuation: bool,
        fit_band: int,
        fit_exclude: float,
        otf_file_location=None,
    ):
        self.otf_correction = otf_correction
        self.otf_type = otf_type
        self.otf_before_shift = otf_before_shift
        self.attenuation_strength = attenuation_strength
        self.attenuation_fwhm = attenuation_fwhm
        self.use_attenuation = use_attenuation
        self.fit_band = fit_band
        self.fit_exclude = fit_exclude
        self.otf_file_location = otf_file_location

    def as_dict(self) -> dict:
        out_dict = {}
        out_dict["otf_correction"] = self.otf_correction
        out_dict["otf_type"] = self.otf_type
        out_dict["otf_before_shift"] = self.otf_before_shift
        out_dict["attenuation_strength"] = self.attenuation_strength
        out_dict["attenuation_fwhm"] = self.attenuation_fwhm
        out_dict["use_attenuation"] = self.use_attenuation
        out_dict["fit_band"] = self.fit_band
        out_dict["fit_exclude"] = self.fit_exclude
        out_dict["otf_file_location"] = self.otf_file_location

        return out_dict

    @classmethod
    def from_dict(cls, data_dict: dict) -> "SIM_Parameters":
        otf_correction = data_dict["otf_correction"]
        otf_type = data_dict["otf_type"]
        otf_before_shift = data_dict["otf_before_shift"]
        attenuation_strength = data_dict["attenuation_strength"]
        attenuation_fwhm = data_dict["attenuation_fwhm"]
        use_attenuation = data_dict["use_attenuation"]
        fit_band = data_dict["fit_band"]
        fit_exclude = data_dict["fit_exclude"]
        otf_file_location = data_dict["otf_file_location"]
        return SIM_Parameters(
            otf_correction,
            otf_type,
            otf_before_shift,
            attenuation_strength,
            attenuation_fwhm,
            use_attenuation,
            fit_band,
            fit_exclude,
            otf_file_location,
        )

    @classmethod
    def as_descriptive_dict(cls) -> dict:
        out_dict = {}
        out_dict[
            "otf_correction"
        ] = "Correction factor to account for worse than ideal otfs."
        out_dict[
            "otf_type"
        ] = "Otf type. Currently supported: EXPONENTIAL, SPHERICAL, NONE, FILE"
        out_dict[
            "otf_before_shift"
        ] = "Determines if otf is applied before shifting in the reconstruction"
        out_dict["attenuation_strength"] = "Attenuation strength from 0 .. 1"
        out_dict["attenuation_fwhm"] = "Attenuation full width at half maximum in 1/um"
        out_dict["use_attenuation"] = "if true use attenuation in wiener filter"
        out_dict["fit_band"] = "Which band to do the fit in the parameter estimation"
        out_dict[
            "fit_exclude"
        ] = "If set to false fitting is skipped and the values for kx, ky from this file are used"
        out_dict[
            "otf_file_location"
        ] = "If a otf file is used include the path here, else set it to null"

        return out_dict


class Microscope_Parameters:
    def __init__(self, pixel_size, na, wavelength):
        self.pixel_size = pixel_size
        self.na = na
        self.wavelength = wavelength

    def as_dict(self) -> dict:
        out_dict = {}
        out_dict["pixel_size"] = self.pixel_size
        out_dict["na"] = self.na
        out_dict["wavelength"] = self.wavelength

        return out_dict

    @classmethod
    def from_dict(cls, data_dict) -> "Microscope_Parameters":
        pixel_size = data_dict["pixel_size"]
        na = data_dict["na"]
        wavelength = data_dict["wavelength"]

        return Microscope_Parameters(pixel_size, na, wavelength)

    @classmethod
    def as_descriptive_dict(cls) -> dict:
        out_dict = {}
        out_dict["pixel_size"] = "Resulting pixel size of the microscope in nm"
        out_dict["na"] = "NA of the objective"
        out_dict["wavelength"] = "Emission wavelength of the microscope in nm"

        return out_dict


class Filter_Parameters:
    def __init__(self, wiener_filter_parameter, apodization_cutoff, apodization_bend):
        # filter parameters:
        self.wiener_filter_parameter = wiener_filter_parameter
        self.apodization_cutoff = apodization_cutoff  # in units the abbe limit frequency of the microscope? -> 2 means 2 times resolution of widefield
        self.apodization_bend = apodization_bend

    def as_dict(self) -> dict:
        out_dict = {}
        out_dict["wiener_filter_parameter"] = self.wiener_filter_parameter
        out_dict["apodization_cutoff"] = self.apodization_cutoff
        out_dict["apodization_bend"] = self.apodization_bend

        return out_dict

    @classmethod
    def as_descriptive_dict(cls) -> dict:
        out_dict = {}
        out_dict[
            "wiener_filter_parameter"
        ] = "Wiener filter parameter usual value is 0.05"
        out_dict[
            "apodization_cutoff"
        ] = "Apodization cutoff in units the abbe limit frequency of the microscope. Usual value is 2"
        out_dict[
            "apodization_bend"
        ] = "Apodization bend. Exponent of the exponentiation of the otf. Usual value is 0.9"

        return out_dict

    @classmethod
    def from_dict(cls, data_dict: dict) -> "Filter_Parameters":
        wiener_filter_parameter = data_dict["wiener_filter_parameter"]
        apodization_cutoff = data_dict["apodization_cutoff"]
        apodization_bend = data_dict["apodization_bend"]
        return Filter_Parameters(
            wiener_filter_parameter, apodization_cutoff, apodization_bend
        )


class Image_Parameters:
    def __init__(
        self,
        image_shape: Tuple[int, int],
        ordering: str,
        z_slice: int,
        background: float,
    ):
        self.image_shape = image_shape
        self.ordering = ordering
        self.z_slice = z_slice
        self.background = background

    def as_dict(self) -> dict:
        out_dict = {}
        out_dict["image_shape"] = self.image_shape
        out_dict["ordering"] = self.ordering
        out_dict["z_slice"] = self.z_slice
        out_dict["background"] = self.background

        return out_dict

    @classmethod
    def as_descriptive_dict(cls) -> dict:
        out_dict = {}
        out_dict[
            "image_shape"
        ] = "the image shape as list in pixel. For now only square images."
        out_dict[
            "ordering"
        ] = "the ordering of the images: p for phase, z for z, a for angel."
        out_dict[
            "z_slice"
        ] = "which z slice should be reconstructed. Starts to count from 0"
        out_dict[
            "background"
        ] = "Background value that should be subtracted from the raw data before further processing"

        return out_dict

    @classmethod
    def from_dict(cls, data_dict: dict) -> "Image_Parameters":
        image_shape = data_dict["image_shape"]
        ordering = data_dict["ordering"]
        z_slice = data_dict["z_slice"]
        background = data_dict["background"]
        return Image_Parameters(image_shape, ordering, z_slice, background)


class Parameters:
    def __init__(
        self,
        directions: List[Direction],
        microscope_parameters: Microscope_Parameters,
        sim_parameters: SIM_Parameters,
        filter_parameters: Filter_Parameters,
        image_parameters: Image_Parameters,
    ):
        self.directions = directions
        self.microscope_parameters = microscope_parameters
        self.sim_parameters = sim_parameters
        self.filter_parameters = filter_parameters
        self.image_parameters = image_parameters
        self.cycles = 1 / (
            image_parameters.image_shape[0] * microscope_parameters.pixel_size
        )
        self.cutoff = (
            2
            * self.microscope_parameters.na
            * self.microscope_parameters.pixel_size
            / self.microscope_parameters.wavelength
        )
        self.otf_array = None
        self.otf_attenuated_array = None

    def as_dict(self) -> dict:
        out_dict = {}

        out_dict["directions"] = [d.as_dict() for d in self.directions]
        out_dict["microscope_parameters"] = self.microscope_parameters.as_dict()
        out_dict["sim_parameters"] = self.sim_parameters.as_dict()
        out_dict["filter_parameters"] = self.filter_parameters.as_dict()
        out_dict["image_parameters"] = self.image_parameters.as_dict()

        return out_dict

    @classmethod
    def as_descriptive_dict(cls) -> dict:
        out_dict = {}

        out_dict["directions"] = [Direction.as_descriptive_dict() for _ in range(3)]
        out_dict["microscope_parameters"] = Microscope_Parameters.as_descriptive_dict()
        out_dict["sim_parameters"] = SIM_Parameters.as_descriptive_dict()
        out_dict["filter_parameters"] = Filter_Parameters.as_descriptive_dict()
        out_dict["image_parameters"] = Image_Parameters.as_descriptive_dict()

        return out_dict

    @classmethod
    def from_dict(cls, data_dict) -> "Parameters":
        directions = [Direction.from_dict(d) for d in data_dict["directions"]]
        microscope_parameters = Microscope_Parameters.from_dict(
            data_dict["microscope_parameters"]
        )
        sim_parameters = SIM_Parameters.from_dict(data_dict["sim_parameters"])
        filter_parameters = Filter_Parameters.from_dict(data_dict["filter_parameters"])
        image_parameters = Image_Parameters.from_dict(data_dict["image_parameters"])

        return Parameters(
            directions,
            microscope_parameters,
            sim_parameters,
            filter_parameters,
            image_parameters,
        )

    def get_otf(self, kx=None, ky=None, attenuated=None) -> ArrayLike:
        if type(self.otf_array) == type(None):
            self._load_otf()
        if kx == None:
            kx = 0
        if ky == None:
            ky = 0
        if attenuated == None:
            otf = (
                self.otf_attenuated_array
                if self.sim_parameters.use_attenuation
                else self.otf_array
            )
        else:
            otf = self.otf_attenuated_array if attenuated else self.otf_array
        return shift_otf(otf, kx, ky)

    def _load_otf(self) -> None:
        kx, ky = 0, 0
        attenuation_fwhm = (
            self.sim_parameters.attenuation_fwhm
            * self.microscope_parameters.pixel_size
            * 1e-3
        )
        # 1e-3 because the input is in nm but the fwhm is for um
        if self.sim_parameters.otf_type.upper() == "FILE":
            self.otf_array = get_otf_from_file(
                self.sim_parameters.otf_file_location, self.image_parameters.image_shape
            )
        else:
            self.otf_array = get_ideal_otf_array(
                self.image_parameters.image_shape,
                self.cutoff,
                kx,
                ky,
                self.sim_parameters.otf_type.upper(),
                self.sim_parameters.otf_correction,
            )

        self.otf_attenuated_array = self.otf_array * get_attenuation_array(
            self.image_parameters.image_shape,
            self.sim_parameters.attenuation_strength,
            attenuation_fwhm,
            kx,
            ky,
        )
        return



import logging
import scipy
import numpy as np
import numpy.typing as npt
from typing import Optional, Tuple, List
import pyfairsim.utils.tif

import matplotlib.pyplot as plt

APPROX_TYPE = ("EXPONENTIAL", "SPHERICAL", "NONE", "FILE")

logger = logging.getLogger(__name__)


def get_attenuation_array(
    size: Tuple[int],
    strength: float,
    fwhm: float,
    k_x: Optional[float] = 0,
    k_y: Optional[float] = 0,
) -> npt.ArrayLike:
    """
    Calculates 2D attenuation array that can be multiplied element wise to an otf to receive an attenuated otf
    size: Tuple of the dimensions for the output array
    strength: strength factor of the gauss
    fwhm: full width at half maximum of the gauss (the FairSIM fwhm needs to be multiplied by the pixel_size)
    k_x/y: shift of center in x/y direction
    """
    # create array containing relative distances to the center size[0]/2, size[1]/2
    y, x = np.ogrid[0 : size[0], 0 : size[1]]
    # delta x = x - k_x - size_x/2 => delta x _relative = delta x / size_x
    relative_distances = np.sqrt(
        ((x - (size[1] / 2 + k_x)) / size[1]) ** 2
        + ((y - (size[0] / 2 - k_y)) / size[0]) ** 2
    )
    # calculate the attenuation
    attenuation = 1 - strength * np.exp(
        -(np.power(relative_distances, 2)) / (2 * np.power(fwhm / 2.355, 2))
    )
    return np.fft.ifftshift(attenuation)


def get_apotization_array(
    size: Tuple[int], cutoff: float, bend: float
) -> npt.ArrayLike:
    """
    Calculates the 2D apotization array
    cutoff: cutoff of the otf used
    bend: bend exponent
    """
    return np.power(get_ideal_otf_array(size, cutoff), bend)


def get_ideal_otf_array(
    size: Tuple[int],
    cutoff: float,
    k_x: Optional[float] = 0,
    k_y: Optional[float] = 0,
    compensation_type: Optional[str] = "",
    a: Optional[float] = 0.15,
) -> npt.ArrayLike:
    """
    Calculates the 2D otf for an ideal lens system, resolution limited by a circular pupil in fourier place.
    "Joseph W. Goodman, Introduction to Fourier Optics, 3. edition, page 145"
    size: Tuple of the dimensions for the output array
    cutoff: normalized cutoff frequency, cutoff = wavelength / (size[0] * pixel_size * NA)
    k_x/y: shift of otf in x/y direction
    compensation_type: Type to compensate for deviations from the ideal otf, can either be EXPONENTIAL or SPHERICAL anything else is treated as None
    a: parameter a for the compensation
    """

    # create array containing relative distances to the center size[0]/2, size[1]/2 and set all distances > 1 to nan
    y, x = np.ogrid[0 : size[0], 0 : size[1]]
    relative_distances = np.sqrt(
        ((x - (size[1] / 2 + k_x)) / size[1]) ** 2
        + ((y - (size[0] / 2 - k_y)) / size[0]) ** 2
    )
    relative_distances[relative_distances > cutoff] = np.nan
    # re-normalize the distances
    relative_distances /= cutoff
    # calculate a ideal otf from relative distances
    otf = (
        2
        / np.pi
        * (
            np.arccos(relative_distances)
            - relative_distances * np.sqrt(1 - relative_distances ** 2)
        )
    )
    # compensate for deviations
    if compensation_type == APPROX_TYPE[0]:  # EXPONENTIAL
        otf *= np.power(a, relative_distances)
    elif compensation_type == APPROX_TYPE[1]:  # SPHERICAL
        otf *= 1 - ((1 - a) * (1 - 4 * np.power(relative_distances - 0.5, 2)))
    # set out of scope otf to 0
    otf[np.isnan(otf)] = 0
    return np.fft.ifftshift(otf)


def shift_otf(otf_array: npt.ArrayLike, kx: float, ky: float) -> npt.ArrayLike:
    if (kx == 0) and (ky == 0):
        # we dont need to shift so no need to calculate
        return np.copy(otf_array)
    return np.fft.ifftshift(scipy.ndimage.shift(np.fft.fftshift(otf_array), (-ky, kx)))


def plot_img(img, title):
    plt.imshow(np.real(img), cmap="gray")
    plt.title(title)
    plt.show()


def get_otf_from_file(filename: str, size: Tuple[int]) -> npt.ArrayLike:
    otf, metadata = pyfairsim.utils.tif.load_tif(filename)
    otf = np.array(otf[0], dtype=np.complex128)
    # scale the image range from 0 to 1
    minimum, maximum = np.min(otf), np.max(otf)
    otf -= minimum
    otf /= maximum - minimum
    if otf.shape != tuple(size):
        raise ValueError(
            f"Shape of otf file ({otf.shape}) and images ({size}) dont match"
        )
    else:
        return np.fft.fftshift(otf)



import logging
import scipy
import numpy as np
import numpy.typing as npt

from pyfairsim.sim_algorithm.SIM_Parameters import Parameters

logger = logging.getLogger(__name__)

import matplotlib.pyplot as plt


def plot_img(img, title):
    plt.imshow(np.real(img), cmap="gray")
    plt.title(title)
    plt.show()


class Wiener_Filter:
    def __init__(self, parameters: Parameters) -> None:
        self.parameters = parameters
        self._update_cache()

    def _update_cache(self) -> None:
        """
        Setup the Wiener filter. This (re)initiates the cached
        filter values. Has to be called if the OTF or the shift parameters change.
        """
        h, w = (
            2 * self.parameters.image_parameters.image_shape[0],
            2 * self.parameters.image_parameters.image_shape[1],
        )
        self.wiener_denominator = np.zeros((h, w), dtype=np.complex128)
        self.otf_array = np.zeros((h, w), dtype=np.complex128)
        self.otf_array[
            h // 4 : (3 * h) // 4, w // 4 : (3 * w) // 4
        ] += np.fft.ifftshift(self.parameters.get_otf(attenuated=False))
        self.attenuated_otf_array = np.zeros((h, w), dtype=np.complex128)
        self.attenuated_otf_array[
            h // 4 : (3 * h) // 4, w // 4 : (3 * w) // 4
        ] += np.fft.ifftshift(self.parameters.get_otf(attenuated=True))

        # loop directions, bands
        for direction in range(len(self.parameters.directions)):
            for band in range(self.parameters.directions[direction].number_bands):
                self._add_wiener_denominator(
                    direction, band, self.parameters.sim_parameters.use_attenuation
                )

    def _add_wiener_denominator(
        self,
        direction_index: int,
        band_index: int,
        use_attenuation: bool,
        return_array=False,
    ) -> None:
        """
        Add OTF^2, for band and direction, to a vector.
        @param vec input vector
        @param direction_index Direction
        @param band_index Band"""
        direction = self.parameters.directions[direction_index]
        kx = direction.px * band_index
        ky = direction.py * band_index

        otf_array_plus = scipy.ndimage.shift(self.otf_array, (ky, kx))
        otf_array_minus = scipy.ndimage.shift(self.otf_array, (-ky, -kx))
        if use_attenuation:
            otf_array_plus *= scipy.ndimage.shift(self.attenuated_otf_array, (ky, kx))
            otf_array_minus *= scipy.ndimage.shift(
                self.attenuated_otf_array, (-ky, -kx)
            )
        else:
            otf_array_plus = np.abs(otf_array_plus) ** 2
            otf_array_minus = np.abs(otf_array_minus) ** 2

        otf_array_plus = np.fft.ifftshift(otf_array_plus)
        otf_array_minus = np.fft.ifftshift(otf_array_minus)
        if not return_array:
            self.wiener_denominator += otf_array_minus + otf_array_plus
        else:
            return otf_array_minus + otf_array_plus

    def get_denominator(self, wiener_parameter: float) -> npt.ArrayLike:
        """
        Returns reciproc Wiener denominator.
        'Reciproc' means the vector can directly be multiplied to spectrum.
        """
        ret = np.copy(self.wiener_denominator)
        return 1 / (ret + wiener_parameter ** 2)

    def get_widefield_denominator(self, wiener_parameter: float) -> npt.ArrayLike:
        h, w = (
            2 * self.parameters.image_parameters.image_shape[0],
            2 * self.parameters.image_parameters.image_shape[1],
        )
        ret = np.zeros((h, w), dtype=np.complex128)
        ret += self._add_wiener_denominator(
            0, 0, use_attenuation=False, return_array=True
        )
        return 1 / (ret + wiener_parameter ** 2)



import logging
from typing import Union, Optional, Dict

logger = logging.getLogger(__name__)

SI_PREFIXES = {
    "Y": 1e24,
    "Z": 1e21,
    "E": 1e18,
    "P": 1e15,
    "T": 1e12,
    "G": 1e9,
    "M": 1e6,
    "k": 1e3,
    "h": 1e2,
    "da": 1e1,
    "d": 1e-1,
    "c": 1e-2,
    "m": 1e-3,
    "u": 1e-6,
    "µ": 1e-6,
    "n": 1e-9,
    "p": 1e-12,
    "f": 1e-15,
    "a": 1e-18,
    "z": 1e-21,
    "y": 1e-24,
}

SI_UNITS = ["m", "kg", "s", "A", "K", "mol", "cd", "1"]


def to_SI(a: Union[float, Dict], unit: Optional[str] = None) -> float:
    """converts the value with unit to the base SI unit"""
    if type(a) == dict:
        unit = a["unit"]
        value = a["value"]
    else:
        value = a
    if "/" in unit:
        nominator, denominator = unit.split("/")
    else:
        nominator = unit
        denominator = ""
    # resolve nominator
    value *= parse_unit_chain(nominator)
    value /= parse_unit_chain(denominator)
    return value


def parse_unit_chain(chain: str) -> float:
    """parses a string of space seperated units.
    returns the factor for conversion to base units."""
    if len(chain) == 0:
        return 1
    conv_factor = 1
    for single_unit in chain.split(
        " "
    ):  # single units need to be spaced so mK /= m K has a clear meaning
        if single_unit in SI_UNITS:  # unit already present as base unit
            continue
        elif single_unit[1:] in SI_UNITS:
            conv_factor *= SI_PREFIXES[single_unit[0]]
        elif single_unit[2:] in SI_UNITS:  # for the case of da which has 2 letters
            conv_factor *= SI_PREFIXES[single_unit[0:2]]
        else:
            raise ValueError(
                f"{single_unit} is not in the supported format for a unit."
            )
    return conv_factor



import logging
import os
import math
import numpy as np
import numpy.typing as npt
import PIL
from PIL.TiffTags import TAGS
from typing import List, Tuple, Dict

# from sim_algorithm.SIM_Parameters import SIM_Parameters
SIM_Parameters = "SIM_Parameters"

logger = logging.getLogger(__name__)


def find_tifs(destination: str) -> List[str]:
    files = []
    for path in os.walk(destination):
        for filename in path[2]:
            if filename.endswith(".tif") or filename.endswith(".tiff"):
                files.append(path[0] + "\\" + filename)
    return files


def tif_basename(filename: str) -> str:
    if filename.endswith(".tiff"):
        return filename[:-5]
    elif filename.endswith(".tif"):
        return filename[:-4]
    else:
        raise ValueError(f"file {filename} dosent end with .tif/.tiff")


def load_tif(image_path: str) -> Tuple[List, Dict]:
    # load image and metadata from file
    with PIL.Image.open(image_path) as tif:
        # assinge the human readable tags to the meta data dict
        meta_data = {
            TAGS[key]: tif.tag[key] for key in tif.tag.keys() if key in TAGS.keys()
        }

        # load images and save them as np arrays in a list
        images = []
        for image_count in range(tif.n_frames):
            tif.seek(image_count)
            images.append(np.array(tif))

        return images, meta_data


def save_tif(image: npt.ArrayLike, path: str) -> None:
    im = PIL.Image.fromarray(image)
    im.save(path)


def order_images(
    images, number_dirs: int, number_phases: int, ordering: str, z_slice: int
):
    length = len(images)
    if (length % (number_dirs * number_phases)) != 0:
        logger.warning(f"the provieded imagestack has missing images.")
    ordering = ordering
    ordered_images = []
    z_slice = z_slice  # 2 for beads, 4 for Lsec
    if ordering == "pza":
        t = length // number_dirs
        for angle in range(number_dirs):
            for phase in range(number_phases):
                ordered_images.append(
                    images[angle * t + phase + z_slice * number_phases]
                )
    elif ordering == "paz":  # this is what we want
        ordered_images = images
    elif ordering == "apz":
        raise NotImplementedError
    elif ordering == "azp":
        raise NotImplementedError
    elif ordering == "zap":
        raise NotImplementedError
    elif ordering == "zpa":
        raise NotImplementedError
    else:
        raise NotImplementedError(f"ordering {ordering} not supported.")
    return ordered_images


def pad_image(image: npt.ArrayLike) -> npt.ArrayLike:
    """
    takes a rectangular image and transforms it into a quadratic
    image with dimensions that are a power of 2 for faster fft.
    @param image numpy array of the image that needs padding"""
    pass  # ToDo


def fade_border_cos(image: np.array, pixels: int) -> np.array:
    """
    Fades borders (sizes px) of the input to zero.
    Done by multiplying sin^2(x) mapped [0..px] to [pi/2..0].
    Good step before zero-padding data for high-res FFT.
    @param image numpy array to fade borders
    @param pixels Size of faded region in pixels
    """
    h, w = image.shape
    fac = 1.0 / pixels * math.pi / 2

    # vertical
    for x in range(w):
        # top
        for y in range(pixels):
            image[y][x] *= math.pow(math.sin(y * fac), 2)
        # bottom
        for y in range(h - pixels, h):
            image[y][x] *= math.pow(math.sin((h - y - 1) * fac), 2)
    # horizontal
    for y in range(h):
        # left
        for x in range(pixels):
            image[y][x] *= math.pow(math.sin(x * fac), 2)
        # right
        for x in range(w - pixels, w):
            image[y][x] *= math.pow(math.sin((w - x - 1) * fac), 2)

    return image


def fft_images(images, number_directions: int, number_phases: int, background: float):
    """
    takes a flat list/array, removes background and fade borders with a cos^2 and then ffts the image.
    returns a array with angle in the first component and phase in the second
    """
    raw_images = np.zeros(
        (
            number_directions,
            number_phases,
            images[0].shape[0],
            images[0].shape[1],
        ),
        dtype=complex,
    )

    for angle in range(number_directions):
        for phase in range(number_phases):
            # select current image
            tmp = images[angle * number_phases + phase]
            

            # convert to float to avoid overflow when substracting
            tmp = tmp.astype(float, copy=True)
            # substract background
            tmp -= background
            tmp[tmp < 0] = 0
            #print(f" input flt {angle} {phase} {np.sum(tmp * tmp)}")

            # apply simple windowing
            tmp = fade_border_cos(tmp, 15)
            #print(f"before fft {angle} {phase} {np.sum(tmp * tmp)}")

            # fft the image
            tmp = np.fft.fft2(tmp, norm="backward")
            #print(f"after fft {angle} {phase} {np.sum(np.abs(tmp * tmp))}")

            # save the image
            raw_images[angle][phase] = tmp

    return raw_images



from setuptools import setup, find_packages

VERSION = "0.0.2"
DESCRIPTION = "A package to run sim reconstructions including parameter estimation."
AUTHOR = "Jakob Wessendorf"

with open("README.md", "r") as f:
    LONG_DESCRIPTION = f.read()

def get_packages():
    print(find_packages())
    return find_packages()

setup(
    name="pyfairsim",
    version=VERSION, 
    author=AUTHOR,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=get_packages(),
    license="",
    url="",
    install_requires=["numpy", "matplotlib", "scipy"],
    python_requires=">=3",
    keywords=["sim", "reconstruction", "parameter estimation"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Operating System :: Microsoft :: Windows",
    ],
)
#ToDo at some point setup.py should be replaced by pyproject.toml


import json
import os

from pyfairsim.sim_algorithm.SIM_Parameters import Parameters


def load_settings(path: str) -> Parameters:
    # try to locate parameter file
    # first check if the path was explicitly given
    if path.endswith(".json"):
        with open(path, encoding="utf8") as f:
            settings = json.load(f)
        return Parameters.from_dict(settings)
    # next check for either image_name.json or parameters.json in the same directory as the image file
    else:
        if path.endswith(".tif"):
            base_file = path[:-4]
        elif path.endswith(".tiff"):
            base_file = path[:-5]
        else:
            raise ValueError(
                f"Unsupported  path: {path}, must either end with .json or .tif/.tiff"
            )
        if os.path.exists(base_file + ".json"):
            with open(path, encoding="utf8") as f:
                settings = json.load(f)
            return Parameters.from_dict(settings)

    base_path = os.path.dirname(path)  # foo/bar/image.tif -> foo/bar
    file_path = os.path.join(base_path, "parameters.json")
    if os.path.exists(file_path):
        with open(file_path, encoding="utf8") as f:
            settings = json.load(f)
        return Parameters.from_dict(settings)
    # load default settings
    default_path = os.path.dirname(os.path.dirname(__file__))
    default_path = os.path.join(default_path, "parameters.json")
    with open(default_path, encoding="utf-8") as f:
        settings = json.load(f)
    return Parameters.from_dict(settings)


def save_settings(settings: Parameters, filename: str, overwrite=False) -> None:
    if os.path.isfile(filename) and not overwrite:
        raise NotImplementedError("File already exists.")
    else:
        with open(filename, "w", encoding="utf-8") as f:
            json.dump(settings.as_dict(), f, indent=4)


def merge_dict(primary, secondary):
    for key, value in secondary.items():
        # if the key dosen't exist create it and continue with the next pair
        if key not in primary:
            primary[key] = value
        # if the value is a dict also merge them
        elif isinstance(value, dict):
            if isinstance(primary[key], dict):
                primary[key] = merge_dict(primary[key], value)
            else:
                raise ValueError(f"{value} and {primary[key]} should be both dicts.")
        # else keep the value stored in primary
    return primary


def create_settings_file(filename):
    settings = Parameters.as_descriptive_dict()
    with open(f"{filename}", "w", encoding="utf-8") as f:
        json.dump(settings, f, indent=4)



import logging
import time

import numpy as np

from pyfairsim.sim_algorithm.SIM_Utils import estimate_parameters, run_reconstruction

from pyfairsim.utils.argparse import parse_args_batch_reconstruction
from pyfairsim.utils.load_settings import load_settings, save_settings
from pyfairsim.utils.tif import (
    load_tif,
    order_images,
    fft_images,
    find_tifs,
    tif_basename,
    save_tif,
)


def main():
    args = parse_args_batch_reconstruction()

    logging.basicConfig(level=args.level)
    global logger
    logger = logging.getLogger()
    logger.info(
        f"Running batch reconstruction on {args.destination} with settings in {args.settings_filename}"
    )

    if args.settings_filename:
        logger.debug(f"loading settings from {args.settings_filename}")
        settings = load_settings(args.settings_filename)
    else:
        logger.warning(f"no settings provided fall back on default settings")
        settings = load_settings(args.destination)

    files_to_reconstruct = find_tifs(args.destination)

    logger.info(f"found {len(files_to_reconstruct)} to reconstruct")

    for filename in files_to_reconstruct:
        try:
            # load images
            logger.debug(f"loading images from {filename}")
            images, _ = load_tif(filename)
            if (
                len(images)
                < len(settings.directions) * settings.directions[0].number_phases
            ):
                logger.error(
                    f"given image stack has to few images n={len(images)} but at least {len(settings.directions) * settings.directions[0].number_phases} are expected"
                )
                raise ValueError("Provided image stack has to few images")

            # order images
            logger.debug(f"order images")
            ordered_images = order_images(
                images,
                len(settings.directions),
                settings.directions[0].number_phases,
                settings.image_parameters.ordering,
                settings.image_parameters.z_slice,
            )

            # fft images
            logger.debug("fft images")
            raw_images = fft_images(
                ordered_images,
                len(settings.directions),
                settings.directions[0].number_phases,
                settings.image_parameters.background,
            )

            # do k-vector and phase estimation
            logger.info(f"run k-vector estimation")
            start = time.time()
            estimate_parameters(settings, raw_images, False)
            stop = time.time()
            logger.info(f"parameter estimation done in {stop-start:.2f} s.")

            save_settings(
                settings, tif_basename(filename) + "_settings.json", overwrite=False
            )

            logger.info(f"run reconstruction")
            start = time.time()
            result = run_reconstruction(settings, raw_images)
            stop = time.time()
            logger.info(f"reconstruction done in {stop-start:.2f} s.")

            save_tif(result[0], tif_basename(filename) + "_reconstructed.tif")

            widefield_filename = (
                tif_basename(args.out_filename) + "_filtered_widefield.tif"
            )
            save_tif(result[1], widefield_filename)

        except Exception as e:
            logger.error(
                f"Error on {filename} skipped this file because: {e.__class__.__name__}"
            )
            logger.error(str(e))
    return


if __name__ == "__main__":
    main()



import argparse


def parse_args_k_estimation() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Running the k-vector estimation on a given image file."
    )
    _register_common_arguments(parser)
    return parser.parse_args()


def parse_args_phase_estimation() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Running the phase estimation on a given image file."
    )
    _register_common_arguments(parser)
    return parser.parse_args()


def parse_args_reconstruction() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Running the sim reconstruction on a given image file."
    )
    _register_common_arguments(parser)
    return parser.parse_args()


def _register_common_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-f",
        "--filename",
        required=True,
        dest="filename",
        type=str,
        help="Input filename",
    )
    parser.add_argument(
        "-o",
        "--out",
        required=False,
        dest="out_filename",
        type=str,
        help="Output filename if empty output is send to the console",
    )
    parser.add_argument(
        "-s",
        "--settings",
        required=False,
        dest="settings_filename",
        type=str,
        help="Settings filename if empty default settings are used",
    )
    parser.add_argument(
        "-d",
        "--debug_level",
        required=False,
        dest="level",
        type=str,
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
        default="INFO",
        help="Sets the log level",
    )


def parse_args_create_settings_file() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Running this script will create a settings file with descriptions instead of values as a template."
    )
    parser.add_argument(
        "-f", dest="filename", type=str, help="Filepath/name for the settings file."
    )
    return parser.parse_args()


def parse_args_batch_reconstruction() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Run a batch parameter estimation and reconstruction on all .tif/.tiff files in the selected folder and all subfolders using one settings file."
    )
    parser.add_argument(
        "-f",
        required=True,
        dest="destination",
        type=str,
        help="Destination folder to run the reconstruction on",
    )
    parser.add_argument(
        "-s",
        "--settings",
        required=False,
        dest="settings_filename",
        type=str,
        help="Settings filename if empty default settings are used",
    )
    parser.add_argument(
        "-d",
        "--debug_level",
        required=False,
        dest="level",
        type=str,
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
        default="INFO",
        help="Sets the log level",
    )
    return parser.parse_args()



import logging
import time

import numpy as np

from pyfairsim.sim_algorithm.SIM_Utils import (
    estimate_absolute_phases,
    estimate_absolute_phases_parallel,
)

from pyfairsim.utils.argparse import parse_args_phase_estimation
from pyfairsim.utils.load_settings import load_settings, save_settings
from pyfairsim.utils.tif import load_tif, order_images, fft_images


def main():
    args = parse_args_phase_estimation()

    logging.basicConfig(level=args.level)
    global logger
    logger = logging.getLogger()

    print(f"Running phase estimation for {args.filename}")

    if args.settings_filename:
        logger.debug(f"loading settings from {args.settings_filename}")
        settings = load_settings(args.settings_filename)
    else:
        logger.warning(f"no settings provided fall back on default settings")
        settings = load_settings(args.filename)

    logger.debug(f"loading images from {args.filename}")
    images, _ = load_tif(args.filename)

    if len(images) < len(settings.directions) * settings.directions[0].number_phases:
        logger.error(
            f"given image stack has to few images n={len(images)} but at least {len(settings.directions) * settings.directions[0].number_phases} are expected"
        )
        return

    logger.debug(f"order images")
    ordered_images = order_images(
        images,
        len(settings.directions),
        settings.directions[0].number_phases,
        settings.image_parameters.ordering,
        settings.image_parameters.z_slice,
    )

    logger.debug("fft images")
    raw_images = fft_images(
        ordered_images,
        len(settings.directions),
        settings.directions[0].number_phases,
        settings.image_parameters.background,
    )

    logger.info(f"run phase estimation")
    start = time.time()
    estimate_absolute_phases_parallel(settings, raw_images)
    stop = time.time()
    logger.info(f"phase estimation done in {stop-start:.2f} s.")

    if args.out_filename:
        save_settings(settings, args.out_filename, overwrite=True)
    else:
        print("Relative phases (difference from phase to phase):")
        for direction in settings.directions:
            phase_offset = direction.phases[-1]
            for phase in direction.phases:
                phase_difference = phase - phase_offset
                # print(phase_difference, end=" converts to ")
                while True:
                    if phase_difference < 0:
                        phase_difference += 2 * np.pi
                    elif phase_difference > 2 * np.pi:
                        phase_difference -= 2 * np.pi
                    else:
                        break
                print(phase_difference, end=" ")
                phase_offset = phase
            print()


if __name__ == "__main__":
    main()
